CON

  _CLKMODE            = XTAL1 + PLL16X
  _XINFREQ            = 5_000_000

  'Clock Frequency Equation
  CLK_FREQ            = ((_CLKMODE - XTAL1) >> 6) * _XINFREQ

  {{
  Propeller Pin Assignment Constants
  }}
  'Z Axis Pin Assignments
  Z_DISABLE           = 0
  Z_STEP              = 1
  Z_DIRECTION         = 2

  'ESTOP And Limit Switches Pin Assignments
  ESTOP_AND_LIMIT     = 3

  'Serial LCD Display Pin Assignment
  SERIAL_LCD          = 4

  'Homing Switch Pin Assignments
  X_HOMING            = 5
  Y_HOMING            = 6
  Z_HOMING            = 7

  'Propeller Memory Card Pin Assignments
  DI                  = 8 'IO0 / DI / MOSI / CD Pin
  DO                  = 9 'IO1 / DO / MISO Pin
  WP                  = 10 'IO2 / WP Pin
  HOLD                = 11 'IO3 / HOLD Pin
  SD_CS               = 12 'CS Pin For MicroSD Card Memory
  FLASH_CS            = 13 'CS Pin For Flash Memory
  SRAM_CS             = 14 'CS Pin For SRAM Memory
  CLK                 = 15 'CLK Pin

  'Daughterboard Pin Assignments
  PLATFORM_HEATER     = 16
  EXTRUDER_HEATER     = 17
  EXTRUDER_FAN        = 18

  'E Axis Pin Assignments  
  E_DIRECTION         = 19
  E_STEP              = 20
  E_DISABLE           = 21

  'X Axis Pin Assignments  
  X_DIRECTION         = 22
  X_STEP              = 23
  X_DISABLE           = 24
  
  'Y Axis Pin Assignments
  Y_DISABLE           = 25
  Y_STEP              = 26  
  Y_DIRECTION         = 27

  'I2C Pin Assignments
  I2C_SCL             = 28
  I2C_SDA             = 29

  'Full Duplex Serial Pin Assignments
  TX                  = 30
  RX                  = 31

  {{
  Additional Constants
  }}
  'X Axis Driver Constants
  X_PULSE_WIDTH       = CLK_FREQ / 1_000_000 '= 80
  X_MIN_SPEED         = CLK_FREQ / 5_714 '= 14_000 Rounded
  X_MAX_SPEED         = CLK_FREQ / 16_000 '= 5_000
  X_RAMP_INC_DEC      = CLK_FREQ / 26_666_666 '= 3 Rounded
  X_INC_DIR           = 0 '= Counter-clockwise rotation
  X_DEC_DIR           = 1 '= Clockwise rotation
  X_STEPS_PER_REV     = 2_000 '= 200 Full steps * 10 microsteps
  X_STEPS_PER_METER   = 55_555
  X_RAMPING_STEPS     = (X_MIN_SPEED - X_MAX_SPEED) / X_RAMP_INC_DEC
  X_TOTAL_RAMPING_STEPS = X_RAMPING_STEPS * 2
  X_MAX_STEPS         = 16_933 '12 inches of linear travel
  X_STEPS_PER_MILLIMETER = 55.55555555555555555555555555555

  'Y Axis Driver Constants
  Y_PULSE_WIDTH       = CLK_FREQ / 1_000_000 '= 80
  Y_MIN_SPEED         = CLK_FREQ / 5_714 '= 14_000 Rounded
  Y_MAX_SPEED         = CLK_FREQ / 16_000 '= 5_000
  Y_RAMP_INC_DEC      = CLK_FREQ / 26_666_666 '= 3 Rounded
  Y_INC_DIR           = 1 '= Counter-clockwise rotation
  Y_DEC_DIR           = 0 '= Clockwise rotation
  Y_STEPS_PER_REV     = 2_000 '= 200 Full steps * 10 microsteps
  Y_STEPS_PER_METER   = 55_555  
  Y_RAMPING_STEPS     = (Y_MIN_SPEED - Y_MAX_SPEED) / Y_RAMP_INC_DEC
  Y_TOTAL_RAMPING_STEPS = Y_RAMPING_STEPS * 2
  Y_MAX_STEPS         = 16_933 '12 inches of linear travel
  Y_STEPS_PER_MILLIMETER = 55.55555555555555555555555555555

  'Full Duplex Serial Driver Constants  
  COMM_MODE           = 0
  COMM_BAUD_RATE      = 115_200

  'FAT Engine Assignment for unused pins or RTC. Please refer to the
  'FAT16/FAT32 Full File System Driver Documentation for proper usage
  NEG_ONE             = -1
  
  'Teacup port
  'Utilized in the creation of a Teacup decfloat structure or the
  'ProcessCoordinate method as it applies to this file
  DECFLOAT_EXPONENT_MAX = 3
  DECFLOAT_MANTISSA_MAX = 4_289_967

VAR

  'The following variable holds the G-Code file line characters.  It has been
  'established that the G-Code file used as a guide in this code, had a maximum
  'file line character count of 72.  To be on the safe side, the maximum file
  'line size has been increased to 100. 
  BYTE bChBuffer[100]

  'The following variables are intended to store G-Code into various fields.
  'For more information about these fields, please refer to RepRap G Code
  'Fields at: http://reprap.org/wiki/G-code .  Additionally, these are the
  'only known fields to be supported by the KISSlicer.  Please note that the
  'maximum character count may change in the future, because I am not exactly
  'sure on bFieldElement_8 maximum character count, but support for 7 characters
  'should be sufficient.  The maximum character count should support a build
  'volume area of 999.99mm(X) X 999.99mm(Y) X 999.99mm(Z).  
  BYTE bFieldElement_0[2] 'G Field
  BYTE bFieldElement_1[3] 'M Field
  BYTE bFieldElement_2[1] 'T Field
  BYTE bFieldElement_3[3] 'S Field
  BYTE bFieldElement_4[7] 'X Field - Floating Point
  BYTE bFieldElement_5[7] 'Y Field - Floating Point
  BYTE bFieldElement_6[7] 'Z Field - Floating Point
  BYTE bFieldElement_7[7] 'F Field - Floating Point
  'BYTE bFieldElement_8[7] 'E Field - Floating Point Temporary comment
  BYTE bFieldElement_8[8] 'E Field - Floating Point

  BYTE bValidFields[9]

  'The following G-Code commands are currently the only known (by me) commands
  'supported by the KISSlicer, therefore these are the only G-Code commands
  'that I will currently support.
  {
        G1: Controlled Move
              G Field - bFieldElement_0, X Field - bFieldElement_4,
              Y Field - bFieldElement_5, Z Field - bFieldElement_6,
              F Field - bFieldElement_7, and E Field - bFieldElement_8
              filled during line parse 
  
        G20: Set Measurement Units To Inches
              G Field - bFieldElement_0 filled during line parse
        
        G21: Set Measurement Units To Millimeters
              G Field - bFieldElement_0 filled during line parse
        
        G28: Home Axes
              G Field - bFieldElement_0 filled during line parse
        
        G90: Set Absolute Positioning
              G Field - bFieldElement_0 filled during line parse
        
        G91: Set Relative Positioning
              G Field - bFieldElement_0 filled during line parse
        
        M82: Set E Codes Absolute
              M Field - bFieldElement_1 filled during line parse
        
        M83: Set E Codes Relative
              M Field - bFieldElement_1 filled during line parse
        
        M104: Set Extruder Temperature With No Wait
              M Field - bFieldElement_1 and S Field - bFieldElement_3
              filled during line parse
  
        M106: Fan On
              M Field - bFieldElement_1 filled during line parse
  
        M109: Set Extruder Temperature And Wait
              M Field - bFieldElement_1 and S Field - bFieldElement_3
              filled during line parse
  
        M107: Fan Off
              M Field - bFieldElement_1 filled during line parse
  
        M140: Set Bed Temperature
              M Field - bFieldElement_1 and S Field - bFieldElement_3
              filled during line parse
  
        T0: Select Tool With No Fields
              T Field - bFieldElement_2 filled during line parse   
  }

  'The following variables are used to calculate syncronized stepper movement
  '''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
  LONG lMoveNow

  BYTE bX_Direction
  LONG lX_MoveComplete
  LONG lX_RampingSteps
  LONG lX_RunningSteps
  LONG lX_Stack[100]
  
  BYTE bY_Direction
  LONG lY_MoveComplete  
  LONG lY_RampingSteps
  LONG lY_RunningSteps
  LONG lY_Stack[100]
  ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
  
  'The following variables will be used to calculate the distance required
  'for targeted moves of the various axes
  '''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''  
  LONG lPreviousX_Location
  LONG lPreviousY_Location
  LONG lPreviousZ_Location
  LONG lPreviousE_Location
  ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''

  'The following three variables are equivalent to a decfloat structure
  'in Teacup, with the added wMultiplicand and is meant to replace the
  'read_digit pointer used extensively in gcode_parse.c  
  '''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''  
  LONG lMantissa 
  BYTE bExponent
  BYTE bSign
  WORD wMultiplicand
  ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''' 

  'The following array is equivalent to the powers constant array in Teacup
  'and used during processing of decfloat_to_int or the ProcessCoordinate
  'method within this file
  '''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
  LONG lPowers[10]
  '''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''

OBJ

  SDCard:       "SD-MMC_FATEngine.spin"
  Memory:       "SPI Memory Driver"
  Terminal:     "FullDuplexSerial"
  Expander:     "SimpleMCP23008"
 
PUB Main

  lPowers[0] := 1
  lPowers[1] := 10
  lPowers[2] := 100
  lPowers[3] := 1_000
  lPowers[4] := 10_000
  lPowers[5] := 100_000
  lPowers[6] := 1_000_000
  lPowers[7] := 10_000_000
  lPowers[8] := 100_000_000
  lPowers[9] := 1_000_000_000

  'Start FullDuplexSerial
  Terminal.Start(RX, TX, COMM_MODE, COMM_BAUD_RATE)
  WAITCNT(CNT + (1 * CLKFREQ))

  'Start FATEngine
  SDCard.fatEngineStart(DO, CLK, DI, SD_CS, NEG_ONE, NEG_ONE, NEG_ONE, NEG_ONE, NEG_ONE)

  'Initialize the MCP23008
  Expander.Initialize

'Success
{
  'Quick test for the MCP23008
}  'Test_MCP23008

'Success
{
  COGNEW(Move_X_Stage, @lX_Stack)
  COGNEW(Move_Y_Stage, @lY_Stack)
  
  WAITCNT(CNT + (3 * CLKFREQ))

  TestStageSyncronization
}

  'Parse the G-Code file MINIMUM.TXT located on SD card
  Parse

PUB TestStageSyncronization | XRampingSteps, XRunningSteps, YRampingSteps, YRunningSteps

  lX_MoveComplete := TRUE
  lY_MoveComplete := TRUE

  UpdateX_RampingAndRunningSteps(X_MAX_STEPS)
  bX_Direction := X_INC_DIR  
  
  UpdateY_RampingAndRunningSteps(Y_MAX_STEPS)
  bY_Direction := Y_INC_DIR

  lMoveNow := TRUE

  REPEAT
  WHILE(lX_MoveComplete == FALSE AND lY_MoveComplete == FALSE)

  lX_MoveComplete := FALSE
  lY_MoveComplete := FALSE    

  WAITCNT(CNT + (5 * CLKFREQ))

  UpdateX_RampingAndRunningSteps(X_MAX_STEPS)
  bX_Direction := X_DEC_DIR  
  
  UpdateY_RampingAndRunningSteps(Y_MAX_STEPS)
  bY_Direction := Y_DEC_DIR

  lMoveNow := TRUE

PUB UpdateX_RampingAndRunningSteps(TotalSteps)

  'If maximum speed can be obtained, determine the actual number of
  'ramping and running steps.
  IF(TotalSteps > X_TOTAL_RAMPING_STEPS)

    'Copy the global amount of ramping steps
    lX_RampingSteps := X_RAMPING_STEPS

    'Determine the number of full speed steps.  
    lX_RunningSteps := TotalSteps - X_TOTAL_RAMPING_STEPS    

   'Else ramp upto the half-way point and then ramp down.    
  ELSE
  
    lX_RampingSteps := TotalSteps / 2
    lX_RunningSteps := TotalSteps // 2

PUB UpdateY_RampingAndRunningSteps(TotalSteps)

  'If maximum speed can be obtained, determine the actual number of
  'ramping and running steps.
  IF(TotalSteps > Y_TOTAL_RAMPING_STEPS)

    'Copy the global amount of ramping steps
    lY_RampingSteps := Y_RAMPING_STEPS

    'Determine the number of full speed steps.  
    lY_RunningSteps := TotalSteps - Y_TOTAL_RAMPING_STEPS    

   'Else ramp upto the half-way point and then ramp down.    
  ELSE
  
    lY_RampingSteps := TotalSteps / 2
    lY_RunningSteps := TotalSteps // 2

PUB Move_X_Stage | Counter, StartStopSpeed

  DIRA[X_DIRECTION] := 1

  REPEAT

    IF(lMoveNow == TRUE)

      lMoveNow := FALSE

      StartStopSpeed := X_MIN_SPEED

      'Set up the CTRMODE of Counter A for NCO/PWM single-ended.
      CTRA[30..26] := %00100

      'Set the output pin for Counter A.
      CTRA[5..0] := X_STEP

      'Set the value to be added to PHSA with every clock cycle.
      FRQA := 1

      'Set APIN as an output.
      DIRA[X_STEP] := 1

      'Set the OUTA register to match the desired direction of rotation.
      OUTA[X_DIRECTION] := bX_Direction

      'Get the current System Counter value.
      Counter := CNT

      'Ramp up the stepper motor to maximum speed.
      REPEAT lX_RampingSteps
  
        'Send out a high pulse on the step pin for the desired duration.
        PHSA := -X_PULSE_WIDTH

        'Wait for a specified period of time before sending another
        'high pulse to the step pin.
        WAITCNT(Counter += StartStopSpeed -= X_RAMP_INC_DEC)

      'Maintain maximum speed    
      REPEAT lX_RunningSteps
   
        'Send out a high pulse on the step pin for the desired duration.
        PHSA := -X_PULSE_WIDTH

        'Wait for a specified period of time before sending another
        'high pulse to the step pin.
        WAITCNT(Counter += StartStopSpeed)

      'Ramp down the stepper motor and come to a stop. 
      REPEAT lX_RampingSteps
  
        'Send out a high pulse on the step pin for the desired duration.
        PHSA := -X_PULSE_WIDTH

        'Wait for a specified period of time before sending another
        'high pulse to the step pin.
        WAITCNT(Counter += StartStopSpeed += X_RAMP_INC_DEC)

      lX_MoveComplete := TRUE

PUB Move_Y_Stage | Counter, StartStopSpeed

  DIRA[Y_DIRECTION] := 1

  REPEAT

    IF(lMoveNow == TRUE)

      lMoveNow := FALSE

      StartStopSpeed := Y_MIN_SPEED

      'Set up the CTRMODE of Counter A for NCO/PWM single-ended.
      CTRA[30..26] := %00100

      'Set the output pin for Counter A.
      CTRA[5..0] := Y_STEP

      'Set the value to be added to PHSA with every clock cycle.
      FRQA := 1

      'Set APIN as an output.
      DIRA[Y_STEP] := 1

      'Set the OUTA register to match the desired direction of rotation.
      OUTA[Y_DIRECTION] := bY_Direction

      'Get the current System Counter value.
      Counter := CNT

      'Ramp up the stepper motor to maximum speed.
      REPEAT lY_RampingSteps
  
        'Send out a high pulse on the step pin for the desired duration.
        PHSA := -Y_PULSE_WIDTH

        'Wait for a specified period of time before sending another
        'high pulse to the step pin.
        WAITCNT(Counter += StartStopSpeed -= Y_RAMP_INC_DEC)

      'Maintain maximum speed    
      REPEAT lY_RunningSteps
   
        'Send out a high pulse on the step pin for the desired duration.
        PHSA := -Y_PULSE_WIDTH

        'Wait for a specified period of time before sending another
        'high pulse to the step pin.
        WAITCNT(Counter += StartStopSpeed)

      'Ramp down the stepper motor and come to a stop. 
      REPEAT lY_RampingSteps
  
        'Send out a high pulse on the step pin for the desired duration.
        PHSA := -Y_PULSE_WIDTH

        'Wait for a specified period of time before sending another
        'high pulse to the step pin.
        WAITCNT(Counter += StartStopSpeed += Y_RAMP_INC_DEC)

      lY_MoveComplete := TRUE
    
PUB Test_MCP23008 | Value

  'GP0 Menu Button
  'GP1 Select Button
  'GP2 Scroll Button
  'GP3 Pause Button
  'GP4 Back Button
  'GP5 Exit Button
  'GP6 Yellow LED
  'GP7 Red LED
    
  Expander.Configure_IODIR_Register(%00111111)  'Configure the IODIR register to 2 outputs and 6 inputs
  Expander.Configure_GPIO_Register(%10000000) '0 is low, GP7, OUTPUT - RED LED

  REPEAT

    Value := Expander.GetPinStates

      IF(Value == %10010000)

        Expander.Configure_GPIO_Register(%11000000)

      ELSE

        Expander.Configure_GPIO_Register(%10000000) '0 is low, GP7, OUTPUT - RED LED

PUB Parse | Mantissa  

  'Mount Partition
  SDCard.mountPartition(0)

  'Open G-Code File "MINIMUM.TXT" for reading
  SDCard.openFile(STRING("MINIMUM.TXT"), "R")

  REPEAT

    'Fill bChBuffer with 0s
    BYTEFILL(@bChBuffer, 0, 100)

    'Fill bValidFields with 0s
    BYTEFILL(@bValidFields, 0, 9)

    'Grab file lines that are not comments or blank
    GetNextValidFileLine

    IF(bChBuffer[0] == 0)
      QUIT

    'Fill in the various field elements
    FillFieldElements
    
    Terminal.Str(@bChBuffer)        

    'Display the various field seperations
    Terminal.Str(STRING("G="))

    IF(bFieldElement_0 == 0)
      Terminal.Str(STRING("NULL"))

    ELSE
      Terminal.Str(@bFieldElement_0)
    
    Terminal.Tx(32) {Space}

    Terminal.Str(STRING("M="))

    IF(bFieldElement_1 == 0)
      Terminal.Str(STRING("NULL"))

    ELSE
      Terminal.Str(@bFieldElement_1)
    
    Terminal.Tx(32) {Space}

    Terminal.Str(STRING("T="))

    IF(bFieldElement_2 == 0)
      Terminal.Str(STRING("NULL"))

    ELSE
      Terminal.Str(@bFieldElement_2)
    
    Terminal.Tx(32) {Space}

    Terminal.Str(STRING("S="))

    IF(bFieldElement_3 == 0)
      Terminal.Str(STRING("NULL"))

    ELSE
      Terminal.Str(@bFieldElement_3)
    
    Terminal.Tx(32) {Space}

    Terminal.Str(STRING("X="))

    IF(bFieldElement_4 == 0)
      Terminal.Str(STRING("NULL"))

    ELSE
      Terminal.Str(@bFieldElement_4)
      'MySteps := ConvertDifferenceToSteps(@bFieldElement_4)
      'lPreviousX_Location := @bFieldElement_4
      Terminal.Tx(32) {Space}
      
      Mantissa := ProcessCoordinate(@bFieldElement_4)
      Terminal.Dec(Mantissa)
    
    Terminal.Tx(32) {Space}

    Terminal.Str(STRING("Y="))

    IF(bFieldElement_5 == 0)
      Terminal.Str(STRING("NULL"))

    ELSE
      Terminal.Str(@bFieldElement_5)
    
    Terminal.Tx(32) {Space}

    Terminal.Str(STRING("Z="))

    IF(bFieldElement_6 == 0)
      Terminal.Str(STRING("NULL"))

    ELSE
      Terminal.Str(@bFieldElement_6)
    
    Terminal.Tx(32) {Space}

    Terminal.Str(STRING("F="))

    IF(bFieldElement_7 == 0)
      Terminal.Str(STRING("NULL"))

    ELSE
      Terminal.Str(@bFieldElement_7)
    
    Terminal.Tx(32) {Space}

    Terminal.Str(STRING("E="))

    IF(bFieldElement_8 == 0)
      Terminal.Str(STRING("NULL"))

    ELSE
      Terminal.Str(@bFieldElement_8)
    
    Terminal.Tx(32) {Space}

    'Seperate G-Code field values line by line
    Terminal.Tx(13) {Carriage Return}

    IF STRCOMP(bChBuffer + 0, STRING("M107"))
      Terminal.Str(STRING("Fan Off"))
      Terminal.Tx(13) {Carriage Return}        

  'Unmount Partition
  SDCard.unmountPartition

  'Stop FATEngine
  SDCard.fatEngineStop

  WAITCNT(CNT + (1 * CLKFREQ))

  'Stop FullDuplexSerial  
  Terminal.Stop

PUB GetNextValidFileLine

  REPEAT

    'Grab a file line as a string  
    SDCard.readString(@bChBuffer, 99)

    'Keep repeating while the following conditions are met,
    'because we don't want or need these lines beginning with
    'these characters.
    IF(bChBuffer[0] <> 59 {; Character} AND bChBuffer[0] <> 13 {Carriage Return} AND bChBuffer[0] <> 10 {Line Feed})
      QUIT

PUB FillFieldElements | Char, Index, Field_Element, ValueIndex

  'Empty the field elements of any possible data
  BYTEFILL(@bFieldElement_0, 0, 2)
  BYTEFILL(@bFieldElement_1, 0, 3)
  BYTEFILL(@bFieldElement_2, 0, 1)
  BYTEFILL(@bFieldElement_3, 0, 3)
  BYTEFILL(@bFieldElement_4, 0, 7)
  BYTEFILL(@bFieldElement_5, 0, 7)
  BYTEFILL(@bFieldElement_6, 0, 7)
  BYTEFILL(@bFieldElement_7, 0, 7)
  'BYTEFILL(@bFieldElement_8, 0, 7) Temporary
  BYTEFILL(@bFieldElement_8, 0, 8)

  Index := 0
  ValueIndex := 0

  'Parse file line into seperate field values.  Various fields will remain empty
  REPEAT

    Char := bChBuffer[Index]

      'Exclude these characters from becoming part of are field entries
      IF(Char == 13 {Carriage Return} OR Char == 10 {Line Feed})
        QUIT

      ELSEIF(Char == 32 {Space})
        Index++
        ValueIndex := 0
        NEXT

      ELSEIF(Char == 71 {G Character})

        'Set current field element
        Field_Element := 0
        bValidFields[0] := 1
        Index++
        NEXT

      ELSEIF(Char == 77 {M Character})

        'Set current field element
        Field_Element := 1
        bValidFields[1] := 1
        Index++
        NEXT

      ELSEIF(Char == 84 {T Character})

        'Set current field element
        Field_Element := 2
        bValidFields[2] := 1
        Index++
        NEXT

      ELSEIF(Char == 83 {S Character})

        'Set current field element
        Field_Element := 3
        bValidFields[3] := 1
        Index++
        NEXT

      ELSEIF(Char == 88 {X Character})

        'Set current field element
        Field_Element := 4
        bValidFields[4] := 1
        Index++
        NEXT

      ELSEIF(Char == 89 {Y Character})

        'Set current field element
        Field_Element := 5
        bValidFields[5] := 1
        Index++
        NEXT

      ELSEIF(Char == 90 {Z Character})

        'Set current field element
        Field_Element := 6
        bValidFields[6] := 1
        Index++
        NEXT

      ELSEIF(Char == 70 {F Character})

        'Set current field element
        Field_Element := 7
        bValidFields[7] := 1
        Index++
        NEXT

      ELSEIF(Char == 69 {E Character})

        'Set current field element
        Field_Element := 8
        bValidFields[8] := 1
        Index++
        NEXT

      ELSE

        IF(Field_Element == 0)

          'G Field element character
          bFieldElement_0[ValueIndex] := Char
          ValueIndex++
          Index++
          NEXT

        ELSEIF(Field_Element == 1)

          'M Field element character
          bFieldElement_1[ValueIndex] := Char
          ValueIndex++
          Index++
          NEXT

        ELSEIF(Field_Element == 2)

          'T Field element character
          bFieldElement_2[ValueIndex] := Char
          ValueIndex++
          Index++
          NEXT

        ELSEIF(Field_Element == 3)

          'S Field element character
          bFieldElement_3[ValueIndex] := Char
          ValueIndex++
          Index++
          NEXT

        ELSEIF(Field_Element == 4)

          'X Field element character
          bFieldElement_4[ValueIndex] := Char
          ValueIndex++
          Index++
          NEXT

        ELSEIF(Field_Element == 5)

          'Y Field element character
          bFieldElement_5[ValueIndex] := Char
          ValueIndex++
          Index++
          NEXT

        ELSEIF(Field_Element == 6)

          'Z Field element character
          bFieldElement_6[ValueIndex] := Char
          ValueIndex++
          Index++
          NEXT

        ELSEIF(Field_Element == 7)

          'F Field element character
          bFieldElement_7[ValueIndex] := Char
          ValueIndex++
          Index++
          NEXT

        ELSEIF(Field_Element == 8)

          'E Field element character
          bFieldElement_8[ValueIndex] := Char
          ValueIndex++
          Index++
          NEXT

PUB ProcessCoordinate(StrPtr) | Index, Char, Mantissa

  wMultiplicand := 1000

  Index := 0

  REPEAT

    Char := BYTE[StrPtr][Index]

    IF(Char => "0" AND Char =< "9")

      IF(bExponent < DECFLOAT_EXPONENT_MAX + 1 AND lMantissa < DECFLOAT_MANTISSA_MAX)

        'this is simply mantissa = (mantissa * 10) + atoi(c) in different clothes
        lMantissa := (lMantissa << 3) + (lMantissa << 1) + (Char - "0")
        'lMantissa := (lMantissa * 10) + LONG[bChar]       

        IF bExponent
        
          bExponent++
          
      Index++       

      NEXT
           
    ELSEIF(Char == "-")

      bSign := 1          
      bExponent := 0
      lMantissa := 0

      Index++
      
      NEXT
          
    ELSEIF(Char == ".")

      IF bExponent == 0
          
        bExponent := 1

      Index++

      NEXT

  WHILE((Char => "0") AND (Char =< "9")) OR (Char == ".") OR (Char == "-")

  Terminal.Dec(lMantissa) 

    'If Exponent = 1, it means we've seen a decimal point with no digits after it,
  'and if Exponent = 2, it means we've seen a decimal point with one digit after it,
  'so it's too high by one if not zero
  IF bExponent

    bExponent--

  'This raises range for mm by factor 1000 and for inches by factor 100.
  'It's a bit expensive, but we should have the time while parsing.
  REPEAT
   
    wMultiplicand /= 10
    bExponent--

  WHILE bExponent AND wMultiplicand // 10 == 0

  lMantissa *= wMultiplicand

  IF bExponent

    lMantissa := (lMantissa + lPowers[bExponent] / 2) / lPowers[bExponent]

  IF bSign

    Mantissa := -lMantissa

  ELSE

    Mantissa := lMantissa

  bSign := 0          
  bExponent := 0
  lMantissa := 0
  wMultiplicand := 0

  RETURN Mantissa

PUB uM_DistanceToStepConversion(Distance, StepsPerMeter)

  RETURN MulDivQR(Distance, StepsPerMeter / 1000000, StepsPerMeter // 1000000, 1000000)

PUB MuldivQR(Multiplicand, Qn, Rn, Divisor) |  Quotient, Remainder, Negative

  {
  \param Multiplicand
  \param Qn ( = multiplier / Divisor )
  \param Rn ( = multiplier % Divisor )
  \param Divisor
  \retuRn rounded result of Multiplicand * multiplier / Divisor
  }

  Quotient := 0
  Remainder := 0
  Negative := 0

  IF Multiplicand < 0
  
    Negative := 1
    Multiplicand := -Multiplicand

  REPEAT WHILE Multiplicand
  
    IF Multiplicand & 1
     
      Quotient += Qn
      Remainder += Rn
      
        IF Remainder >= Divisor
        
          Quotient++
          Remainder -= Divisor
      
    Multiplicand  >>= 1
    Qn <<= 1
    Rn <<= 1
    
    IF Rn >= Divisor
    
      Qn++ 
      Rn -= Divisor  

  'Rounding
  IF Remainder > Divisor / 2
  
    Quotient++

  'Remainder is valid, but not returned
  IF Negative == 1

    RETURN -Quotient

  ELSE

    RETURN Quotient

PUB ApproximateDistance(Dx, Dy) | Minimum, Maximum, Approximate

  {
  \param dx distance in X plane
  \param dy distance in Y plane
  \return 3-part linear approximation of \f$\sqrt{\Delta x^2 + \Delta y^2}\f$
  }

  'If either axis is zero, return the other one.
  IF Dx == 0 OR Dy == 0
  
    RETURN Dx + Dy

  IF Dx < Dy
  
    Minimum := Dx
    Maximum := Dy

  ELSE

    Minimum := Dy
    Maximum := Dx

  Approximate := (Maximum * 1007) + (Minimum * 441)

  IF Maximum < (Minimum << 4)
  
    Approximate -= (Maximum * 40)

  'Add 512 for proper rounding
  RETURN (Approximate + 512) >> 10

PUB ApproximateDistance3D(Dx, Dy, Dz) | Minimum, Medium, Maximum, Approximate

  {
  \param dx distance in X plane
  \param dy distance in Y plane
  \param dz distance in Z plane
  \return 3-part linear approximation of \f$\sqrt{\Delta x^2 + \Delta y^2 + \Delta z^2}\f$
  }

  IF Dx < Dy
  
    Minimum := Dy
    Medium := Dx

  ELSE

    Minimum := Dx
    Medium := Dy

  IF Dz < Minimum
  
    Maximum := Medium
    Medium := Minimum
    Minimum := Dz

  ELSEIF Dz < Medium
  
    Maximum := Medium
    Medium := Dz

  ELSE

    Maximum := Dz

  Approximate := (Maximum * 860) + (Medium * 851) + (Minimum * 520)

  IF Maximum < (Medium << 1)

    Approximate -= (Maximum * 294)

  IF Maximum < (Minimum << 2)

    Approximate -= (Maximum * 113)

  IF Medium < (Minimum << 2 )

    Approximate -= (Medium *  40)

  'Add 512 for proper rounding
  RETURN (Approximate + 512 ) >> 10
{
PUB IntegerSquareRoot(IntSqRt_A)

  {
  \param a find square root of this number
  \return sqrt(a - 1) < returnvalue <= sqrt(a)
  }

  wIntSqRt_B := IntSqRt_A >> 16
  bIntSqRt_C := wIntSqRt_B >> 8
  wIntSqRt_X := 0
  bIntSqRt_Z := 0
  
  bIntSqRt_J := $8

  REPEAT WHILE bIntSqRt_J

    bIntSqRt_Z |= bIntSqRt_J
    bIntSqRt_Y2 := bIntSqRt_Z * bIntSqRt_Z
    
    IF bIntSqRt_Y2 > bIntSqRt_C
    
      bIntSqRt_Z ^= bIntSqRt_J

    bIntSqRt_J >>= 1  
  
  wIntSqRt_X := bIntSqRt_Z << 4   

  wIntSqRt_I := $8

  REPEAT WHILE wIntSqRt_I

    wIntSqRt_X |= wIntSqRt_I
    wIntSqRt_Y2 := wIntSqRt_X * wIntSqRt_X
    
    IF wIntSqRt_Y2 > wIntSqRt_B
    
      wIntSqRt_X ^= wIntSqRt_I

    wIntSqRt_I >>= 1   
  
  wIntSqRt_X <<= 8
  
  wIntSqRt_I := $80

  REPEAT WHILE wIntSqRt_I

    wIntSqRt_X |= wIntSqRt_I
    lIntSqRt_Y2 := wIntSqRt_X * wIntSqRt_X
    
    IF lIntSqRt_Y2 > IntSqRt_A
    
      wIntSqRt_X ^= wIntSqRt_I

    wIntSqRt_I >>= 1  

  RETURN wIntSqRt_X

PUB IntegerInvertedSquareRoot(IntInvSqRt_A)

  bIntInvSqRt_Z := 0
  
  lIntInvSqRt_Q := LONG[($FFFF / IntInvSqRt_A)] << 8

  bIntInvSqRt_I := $80

  REPEAT WHILE bIntInvSqRt_I
    
    bIntInvSqRt_Z |= bIntInvSqRt_I
    wIntInvSqRt_Y := bIntInvSqRt_Z * bIntInvSqRt_Z
    
    IF wIntInvSqRt_Y > (lIntInvSqRt_Q >> 8)
    
      bIntInvSqRt_Z ^= bIntInvSqRt_I
      
    bIntInvSqRt_I >>= 1

  wIntInvSqRt_X := bIntInvSqRt_Z << 4

  wIntInvSqRt_J := $8

  REPEAT WHILE wIntInvSqRt_J
  
    wIntInvSqRt_X |= wIntInvSqRt_J    
    lIntInvSqRt_Y := wIntInvSqRt_X * wIntInvSqRt_X
    
    IF lIntInvSqRt_Y > lIntInvSqRt_Q
    
      wIntInvSqRt_X ^= wIntInvSqRt_J

    wIntInvSqRt_J >>= 1

  RETURN wIntInvSqRt_X
}
DAT

        Tag1 BYTE "M107",0